
/*********************************************************/
/**                                                     **/
/**           tcl.c (c) 1997, Emiel Bruijntjes          **/
/**                                                     **/
/**               Handling tcl-interface.               **/
/**                                                     **/
/*********************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif


#include <stdlib.h>


#include "tcl_procs.h"
#include "misc.h"
#include "session.h"
#include "config.h"

#ifdef NO_TCL

void init_tcl() { return; }

void tcl_eval(char *string, void output(char *, ...)) { return; }
void tcl_evalfile(char *file, void output(char *, ...)) { return; }

#else /* not NO_TCL */

#include <tcl.h>

Tcl_Interp *interp;


void init_tcl() {

  interp = Tcl_CreateInterp();

  tcl_initcommands(interp);
}


void tcl_eval(char *string, void output(char *, ...)) {

  if (Tcl_Eval(interp, string) == TCL_ERROR) {
    output("Error in Tcl-command: %s", interp->result);
  }
}

void tcl_evalfile(char *file, void output(char *, ...)) {

  if (Tcl_EvalFile(interp, file) == TCL_ERROR) {
    output("Error in Tcl-script: %s", interp->result);
  }
}

#endif /* not NO_TCL */


struct scriptlist {
  struct scriptinfo *first;
  struct scriptinfo *tail;
  struct scriptinfo *current;
  int length;
} scriptlist={ NULL , NULL , NULL , 0 };

struct scriptinfo {
  char *script;

  struct scriptinfo *next;
  struct scriptinfo *prev;
};


void addscript(char *script) {

  scriptlist.length++;
  scriptlist.current = (struct scriptinfo *) xmalloc (sizeof(struct scriptinfo));
  if (scriptlist.first == NULL) {
    scriptlist.first = scriptlist.current;
    scriptlist.current->prev = NULL;
  }
  else {
    scriptlist.current->prev = scriptlist.tail;
    scriptlist.tail->next = scriptlist.current;
  }
  scriptlist.current->next = NULL;
  scriptlist.tail = scriptlist.current;

  scriptlist.current->script = (char *) xmalloc (strlen(script));
  strcpy(scriptlist.current->script, script);
}


void delscript() { 

  if (scriptlist.current == NULL) return;

  xfree(scriptlist.current->script);
  scriptlist.length--;
  if ((scriptlist.current == scriptlist.first) && (scriptlist.current == scriptlist.tail)) {
    xfree(scriptlist.current);
    scriptlist.current = NULL;
    scriptlist.tail = NULL;
    scriptlist.first = NULL;
    return;
  }

  if (scriptlist.current->next == NULL) {
    scriptlist.tail = scriptlist.current->prev;
    scriptlist.tail->next = NULL;
    xfree(scriptlist.current);
    scriptlist.current = scriptlist.tail;
    return;
  }

  if (scriptlist.current->prev == NULL) {
    scriptlist.first = scriptlist.current->next;
    scriptlist.first->prev = NULL;
    xfree(scriptlist.current);
    scriptlist.current = scriptlist.first;
    return;
  }

  scriptlist.current->prev->next = scriptlist.current->next;
  scriptlist.current->next->prev = scriptlist.current->prev;
  xfree(scriptlist.current);
  scriptlist.current = scriptlist.first;
  return;
}


int findscript(char *script) {

  for (scriptlist.current = scriptlist.first; scriptlist.current; scriptlist.current = scriptlist.current->next) {
    if (strcmp(script, scriptlist.current->script) == 0) return 1;
  }
  return 0;
}


int firstscript() { return (int)(scriptlist.current = scriptlist.first); }
int nextscript() { return (int)(scriptlist.current = scriptlist.current->next); }

char *currentscript() { return scriptlist.current->script; }

void save_scriptlist(FILE *fp) {

  fwriteint(fp, scriptlist.length);

  if (firstscript()) {
    do
      {
	fwritestr2bin(scriptlist.current->script, fp);
      }
    while (nextscript());
  }
  fwriteint(fp, 834567);
}


int load_scriptlist(FILE *fp, int version) {

  int i;

  if (version < 45) return 1;

  while (firstscript()) delscript();

  freadint(fp, &scriptlist.length);
  for (i=0; i < scriptlist.length; i++) {
    scriptlist.current = (struct scriptinfo *) xmalloc (sizeof(struct scriptinfo));
    if (scriptlist.first == NULL) {
      scriptlist.first = scriptlist.current;
      scriptlist.current->prev = NULL;
    }
    else {
      scriptlist.tail->next = scriptlist.current;
      scriptlist.current->prev = scriptlist.tail;
    }
    scriptlist.current->next = NULL;
    scriptlist.tail = scriptlist.current;
    
    scriptlist.current->script = NULL;
    fgetstrfrombin(&scriptlist.current->script, fp);
  }
  freadint(fp, &i);
  return (i == 834567);
}


void eval_scriptlist() {

  if (firstscript()) {
    do 
      {
	tcl_evalfile(currentscript(), int_err);
      }
    while (nextscript());
  }
}
