
/*******************************************************/
/**                                                   **/
/**          term.c (c) 1997 Emiel Bruijntjes         **/
/**                                                   **/
/**        Handling the console on the terminal       **/
/**                                                   **/
/*******************************************************/

#ifdef SUNOS
#include "sunos.h"
#endif


#include <stdlib.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <errno.h>
#include <ctype.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>


#include "misc.h"
#include "term.h"
#include "config.h"
#include "console.h"
#include "party_console.h"
#include "channel.h"
#include "net.h"
#include "history.h"
#include "cmd_console.h"
#include "irc_console.h"
#include "on.h"


/* Some functions and variables that should be in termcap.h: */
char *tgoto(char *, int, int);
int   tgetent(char *, char *);
int   tgetnum(char *);
char *tgetstr(char *, char **);
int   tputs(const char *, int, int (*outc)(int));

extern char *UB;
extern char *BC;
extern char *UP;
/* end of termcap functions and variables */


extern time_t time_null;

       char  have_terminal=0;

extern char *progname;
extern char  current_nick[10];

static int  terminal_socket_ptr;
static char termcap[1024];             /* A string used to store the termcap information */

/* variables from termcap */

static char *empty_string="";

static char *ring_bell,           /* bl */
            *cursor_motion,       /* cm */
            *bold_mode_on,        /* md */
            *all_attr_off,        /* me */
            *reverse_mode_on,     /* se */
            *reverse_mode_off,    /* so */
            *undersc_mode_off,    /* ue */
            *undersc_mode_on,     /* us */
            *clear_to_eol,        /* ce */
            *key_down_arrow,      /* kd */
            *key_up_arrow,        /* ku */
            *key_left_arrow,      /* kl */
            *key_right_arrow,     /* kr */
            *carriage_return,     /* cr */
            *newline;             /* nl */
    
static int   cols, lines;               /* Number of columns and lines on the terminal */

static int   tty_des;                   /* File descriptor of the terminal */

static struct termios old_modes;         /* Old modes of the terminal */

char **screen_array;
int    mode_line;                       /* The upper black line (index in screen_array) */
int    status_line;                     /* The lower black line (index in screen_array) */


void  restore_cursor();
void  print_input_buffer();

int putchar_x(int c) { fputc(c, stdout); return 1; }


/* Print the mode-line, the black line between the upper window and the IRC-window */
void term_print_mode_line() {


  if (screen_array[mode_line] == NULL) {
    screen_array[mode_line] = (char *) xmalloc (MAX(cols + 1 + strlen(reverse_mode_on) + strlen(reverse_mode_off) + 25,
						    300));
  }

  sprintf(screen_array[mode_line], "%s Mode: %-7s (CTRL + \\)   channel: %s%s%s", reverse_mode_on, 
	  console_mode() == CONSOLE_IRC_MODE ? "IRC" : console_mode() == CONSOLE_COMMAND_MODE ? "Command" : "Party", 
	  current_party_channel(), spaces(cols - 41), reverse_mode_off);

  tputs(tgoto(cursor_motion, 0, mode_line), 1, putchar_x);
  tputs(screen_array[mode_line], 1, putchar_x);
  fflush(stdout);
  restore_cursor();
}


/* Print the status-line, the black line between the IRC-window and input-window */
void term_print_status_line() {

  struct  tm *timeptr;
  time_t  curtime;
  char   *channel;

  if (screen_array[status_line] == NULL) {
    screen_array[status_line] = (char *) xmalloc (MAX(cols + 1 + strlen(reverse_mode_on) + strlen(reverse_mode_off) + 25,
						      300));
  }

  curtime = time_null;
  timeptr = localtime(&curtime);

  if (channellist_findfirst())
    channel = channellist_channelname();
  else
    channel = "";

  sprintf(screen_array[status_line], "%s %d:%02d %c%-9s channel: %s %s [%s] %s", reverse_mode_on, timeptr->tm_hour, 
	  timeptr->tm_min, ischanneloperator(channel, current_nick) ? '@' : ' ', current_nick, channel, 
	  spaces(cols-33-strlen(channel)-strlen(VERSION)), VERSION, reverse_mode_off);
  
  tputs(tgoto(cursor_motion, 0, status_line), 1, putchar_x);
  tputs(screen_array[status_line], 1, putchar_x);
  fflush(stdout);
  restore_cursor();
}


/* Initialize the terminal settings */
void init_terminal() {

  /* Initialize all the terminal settings */

  struct termios new_modes;

  char  *term, *ptr;
  char   bp[1024];
  int    i;

  tty_des = fileno(stdout);             /* Set the filedescriptor of the terminal */

  if ((term = getenv("TERM")) == NULL) {
    fprintf(stderr, "%s: No TERM variable set!\n", progname);
    fprintf(stderr, "%s: You may still run the bot without a console.\n", progname);
    exit(1);
  }

  switch (tgetent(bp, term)) {
  case -1:
    fprintf(stderr, "%s: None of the termcap database files could be opened.\n", progname);
    fprintf(stderr, "%s: You may still run this bot without a console.\n", progname);
    exit(1);
  case  0:
    fprintf(stderr, "%s: Your terminal does not have an entry in the termcap database.\n", progname);
    fprintf(stderr, "%s: You may still run this bot without a console.\n", progname);
    exit(1);
  }
  
  /* Saving the current settings of the terminal */

  if (tcgetattr(tty_des, &old_modes) < 0) {
    switch (errno) {
    case EBADF:
      fprintf(stderr, "%s: %d is an invalid filedescriptor.\n", progname, tty_des);
      exit(1);
    case ENOTTY:
      fprintf(stderr, "%s: File descriptor %d is not associated with a terminal.\n", progname, tty_des);
      exit(1);
    default:
      fprintf(stderr, "%s: Error %d", progname, errno);
      exit(1);
    }
  }
  
  memcpy(&new_modes, &old_modes, sizeof(struct termios));
  
  new_modes.c_lflag &= ~(ICANON | ECHO);  /* cbreak and no echo */
  
  new_modes.c_cc[VMIN] = 1;    /* read() satified after 1 char */
  new_modes.c_cc[VTIME] = 0;   /* No timer */
#ifdef _POSIX_VDISABLE
  new_modes.c_cc[VQUIT] = _POSIX_VDISABLE;
#endif

  if (tcsetattr(tty_des, TCSADRAIN, &new_modes) < 0) {
    switch (errno) {
    case EBADF:
      fprintf(stderr, "%s: %d is an invalid filedescriptor.\n", progname, tty_des);
      exit(1);
    case ENOTTY:
      fprintf(stderr, "%s: File descriptor %d is not associated with a terminal.\n", progname, tty_des);
      exit(1);
    case EINVAL:
      fprintf(stderr, "%s: Error in parameters of tcsetattr()", progname);
      exit(1);
    default:
      fprintf(stderr, "%s: Error %d", progname, errno);
      exit(1);
    }
  }
  
  if ((cols  = tgetnum("co")) == -1) cols  = 80;
  if ((lines = tgetnum("li")) == -1) lines = 24;

  ptr = termcap;

  if (((cursor_motion = tgetstr("cm", &ptr)) == NULL)) {
    fprintf(stderr, "%s: Your terminal does not have the necessary capabilities for the console.\n", progname);
    fprintf(stderr, "%s: You may still run this bot without a console.\n", progname);
    exit(1);
  }

  if ((ring_bell = tgetstr("bl", &ptr)) == NULL) ring_bell = empty_string;
  if ((bold_mode_on = tgetstr("md", &ptr)) == NULL) {
    if ((bold_mode_on = tgetstr("mb", &ptr)) == NULL)
      bold_mode_on = empty_string;
  }
  if ((all_attr_off = tgetstr("me", &ptr)) == NULL) all_attr_off = empty_string;
  if ((reverse_mode_on = tgetstr("so", &ptr)) == NULL) {
    if ((reverse_mode_on = tgetstr("mr", &ptr)) == NULL)
      reverse_mode_on = empty_string;
  }
  if ((reverse_mode_off = tgetstr("se", &ptr)) == NULL) {
    if ((reverse_mode_off = tgetstr("me", &ptr)) == NULL)
      reverse_mode_off = empty_string;
  }
  if ((undersc_mode_on = tgetstr("us", &ptr)) == NULL) undersc_mode_on = empty_string;
  if ((undersc_mode_off = tgetstr("ue", &ptr)) == NULL) undersc_mode_off = empty_string;

  if ((UP = tgetstr("up", &ptr)) == NULL) UP = empty_string;
  if ((BC = tgetstr("bc", &ptr)) == NULL) BC = empty_string;

  if ((clear_to_eol = tgetstr("ce", &ptr)) == NULL) { clear_to_eol = empty_string; }
  if ((carriage_return = tgetstr("cr", &ptr)) == NULL) { carriage_return = "\r"; }
  if ((newline = tgetstr("nl", &ptr)) == NULL) { newline = "\n"; }

  key_up_arrow    = tgetstr("ku", &ptr);
  key_down_arrow  = tgetstr("kd", &ptr);
  key_left_arrow  = tgetstr("kl", &ptr);
  key_right_arrow = tgetstr("kr", &ptr);

  /* initializing the array with strings */
  screen_array = (char **) xmalloc (sizeof(char *) * lines);
  mode_line    = (lines / 2) - 1;
  status_line  = lines - 2;

  for (i=0; i < mode_line; i++) {
    screen_array[i] = (char *) xmalloc (1);
    strcpy(screen_array[i], "");
  }

  for (i = mode_line + 1; i < status_line; i++) {
    screen_array[i] = (char *) xmalloc (1);
    strcpy(screen_array[i], "");
  }
  screen_array[mode_line]   = NULL;
  screen_array[status_line] = NULL;

  term_print_status_line();
  term_print_mode_line();

  term_flush_window(TERM_SNOOP_WINDOW);
  term_flush_window(TERM_IRC_WINDOW);

  print_input_buffer();

  have_terminal = 1;
  if ((terminal_socket_ptr = get_stdin()) == -1) {
    fprintf(stderr, "Could not create the console. (All connections in use)\n");
    stop_terminal();
  }
}


/* Scroll a window */
static void scrollwindow(char window) {

  int i;

  switch (window) {
  case TERM_SNOOP_WINDOW:
    xfree (screen_array[0]);
    for (i=0; i < mode_line - 1; i++) {
      screen_array[i] = screen_array[i+1];
    }
    screen_array[mode_line - 1] = NULL;
    break;
  case TERM_IRC_WINDOW:
    xfree (screen_array[mode_line + 1]);
    for (i = mode_line + 1; i < status_line - 1; i++) {
      screen_array[i] = screen_array[i+1];
    }
    screen_array[status_line - 1] = NULL;
    break;
  }
}


/* Returns the length of the next word. Control characters are not counted */
static int t_strlen_nextword(char *string) {

  int i=0, j=0;

  while ((!isspace(string[i])) && (string[i] != '\0')) {
    if (isprint(string[i])) j++;
    i++;
  }
  return j;
}


#define A_BOLD      1
#define A_UNDERLINE 2
#define A_REVERSE   4


/* Local procedure to print something on one of the two windows. Use term_send2window to print something */
static void send2terminalwindow_(char window, char *string, char attr, int startpos) {

  int   scr_index=0;
  char *lastchar;
  int   cursor, wordlength;

  scrollwindow(window);

  switch (window) {
  case TERM_SNOOP_WINDOW:
    scr_index = mode_line - 1;
    break;
  case TERM_IRC_WINDOW:
    scr_index = status_line - 1;
    break;
  }

  screen_array[scr_index] = (char *) xmalloc (WAYTOBIG);
  strcpy(screen_array[scr_index], spaces(startpos));
  cursor = startpos;
  lastchar = index(screen_array[scr_index], '\0');
  if (attr & A_BOLD)      { strcpy(lastchar, ""); lastchar = index(lastchar, '\0'); }
  if (attr & A_UNDERLINE) { strcpy(lastchar, ""); lastchar = index(lastchar, '\0'); }
  if (attr & A_REVERSE)   { strcpy(lastchar, ""); lastchar = index(lastchar, '\0'); }

  while (1) {
    wordlength = t_strlen_nextword(string);

    if (wordlength < cols - cursor) {
      while (isgraph(*string)) {
	*lastchar = *string;
	lastchar++; string++; cursor++;
      }
      *lastchar = '\0';
    }
    else if (wordlength > cols - startpos) {
      while (isgraph(*string) && cursor < cols) {
	*lastchar = *string;
	lastchar++; string++; cursor++;
      }
      *lastchar = '\0';
      if (cursor >= cols) {
	screen_array[scr_index] = (char *) xrealloc (screen_array[scr_index], strlen(screen_array[scr_index]) + 1);
	send2terminalwindow_(window, string, attr, startpos);
	return;
      }
    }
    else {
      *lastchar = '\0';
      screen_array[scr_index] = (char *) xrealloc (screen_array[scr_index], strlen(screen_array[scr_index]) + 1);
      send2terminalwindow_(window, string, attr, startpos);
      return;
    }
    while (!isgraph(*string)) {
      switch (*string) {
      case ' ':
	if (startpos == 0) startpos = cursor + 1;
	*lastchar = *string;
	lastchar++; string++; cursor++;
	*lastchar = '\0';
	if (cursor >= cols) {
	  while (isspace(*string)) string++;
	  screen_array[scr_index] = (char *) xrealloc (screen_array[scr_index], strlen(screen_array[scr_index]) + 1);
	  send2terminalwindow_(window, string, attr, startpos);
	  return;
	}
	break;
      case '':
	*lastchar = *string;
	lastchar++; string++;
	*lastchar = '\0';
	if (attr & A_BOLD)
	  attr &= ~A_BOLD;
	else
	  attr |= A_BOLD;
	break;
      case '':
	*lastchar = *string;
	lastchar++; string++;
	*lastchar = '\0';
	if (attr & A_REVERSE)
	  attr &= ~A_REVERSE;
	else
	  attr |= A_REVERSE;
	break;
      case '':
	*lastchar = *string;
	lastchar++; string++;
	*lastchar = '\0';
	if (attr & A_UNDERLINE)
	  attr &= ~A_UNDERLINE;
	else
	  attr |= A_UNDERLINE;
	break;
      case '\0':
	*lastchar = *string;
	screen_array[scr_index] = (char *) xrealloc (screen_array[scr_index], strlen(screen_array[scr_index]) + 1);
	return;
      default:
	string++;
      }
    }
  }
}


void term_flush_window(char window) {

  int i=0, j=0, attr=0, lb=0, ub=0;

  switch (window) {
  case TERM_SNOOP_WINDOW:
    lb = 0; 
    ub = mode_line;
    break;
  case TERM_IRC_WINDOW:
    lb = mode_line + 1;
    ub = status_line;
    break;
  }

  tputs(tgoto(cursor_motion, 0, lb), 1, putchar_x);
  for (i=lb; i<ub; i++) {
    tputs(clear_to_eol, 1, putchar_x);
    for (j=0; screen_array[i][j]; j++) {
      switch (screen_array[i][j]) {
      case '':
	if (attr & A_BOLD) {
	  tputs(all_attr_off, 1, putchar_x);
	  attr &= ~A_BOLD;
	} else {
	  tputs(bold_mode_on, 1, putchar_x);
	  attr |= A_BOLD;
	}
	break;
      case '':
	if (attr & A_REVERSE) {
	  tputs(reverse_mode_off, 1, putchar_x);
	  attr &= ~A_REVERSE;
	} else {
	  tputs(reverse_mode_on, 1, putchar_x);
	  attr |= A_REVERSE;
	}
	break;
      case '':
	if (attr & A_UNDERLINE) {
	  tputs(undersc_mode_off, 1, putchar_x);
	  attr &= ~A_UNDERLINE;
	} else {
	  tputs(undersc_mode_on, 1, putchar_x);
	  attr |= A_UNDERLINE;
	}
	break;
      default:
	putchar_x(screen_array[i][j]);
      }
    }
    if (attr) {
      tputs(all_attr_off, 1, putchar_x);
      attr = 0;
    }
    tputs(carriage_return, 1, putchar_x);
    tputs(newline, 1, putchar_x);
  }
  restore_cursor();
  fflush(stdout);
}


void term_send2window(char window, char *message) {

  send2terminalwindow_(window, message, 0, 0);
  term_flush_window(window);
}


void stop_terminal() {
  
  /* Restore the values as they were before init_terminal() */
  int i;
  
  if (tcsetattr(tty_des, TCSADRAIN, &old_modes) < 0) {
    switch (errno) {
    case EBADF:
      fprintf(stderr, "%s: %d is an invalid filedescriptor.\n", progname, tty_des);
      exit(1);
    case ENOTTY:
      fprintf(stderr, "%s: File descriptor %d is not associated with a terminal.\n", progname, tty_des);
      exit(1);
    case EINVAL:
      fprintf(stderr, "%s: Error in parameters of tcsetattr()", progname);
      exit(1);
    default:
      fprintf(stderr, "%s: Error %d", progname, errno);
      exit(1);
    }
  }
  for (i=0; i <= status_line + 1; i++)
    xfree (screen_array[i]);

  xfree(screen_array);

  close_socketptr(terminal_socket_ptr);
  terminal_socket_ptr = -1;
}


void resize_terminal() {

#ifdef TIOCGWINSZ

  int i;
  struct winsize window;

  for (i=0; i <= status_line + 1; i++)
    xfree (screen_array[i]);

  xfree(screen_array);

  if (ioctl(tty_des, TIOCGWINSZ, &window) != 0) {

    if (lines != window.ws_row || cols != window.ws_col) {
      for (i=0; i <= status_line + 1; i++)
	xfree (screen_array[i]);
      
      xfree(screen_array);

      lines = window.ws_row;
      cols  = window.ws_col;
      
      screen_array = (char **) xmalloc (sizeof(char *) * lines);
      mode_line    = (lines / 2) - 1;
      status_line  = lines - 2;
      
      for (i=0; i < mode_line; i++) {
	screen_array[i] = (char *) xmalloc (1);
	strcpy(screen_array[i], "");
      }
      
      for (i = mode_line + 1; i < status_line; i++) {
	screen_array[i] = (char *) xmalloc (1);
	strcpy(screen_array[i], "");
      }
      screen_array[mode_line]   = NULL;
      screen_array[status_line] = NULL;
      
      term_print_status_line();
      term_print_mode_line();
      
      term_flush_window(TERM_SNOOP_WINDOW);
      term_flush_window(TERM_IRC_WINDOW);
      
      print_input_buffer();
    }
  }

#endif
}

/**********************************************************************************/
/**                                                                              **/
/**                              Handling of input                               **/
/**                                                                              **/
/**********************************************************************************/


static char input_buffer[WAYTOBIG]="";
static int  buffer_length=0;
static int  input_cursor=0;


void t_putc(char ch) {

  switch(ch) {
  case '':
    tputs(bold_mode_on, 1, putchar_x);
    tputs("B", 1, putchar_x);
    tputs(all_attr_off, 1, putchar_x);
    break;
  case '':
    tputs(bold_mode_on, 1, putchar_x);
    tputs("V", 1, putchar_x);
    tputs(all_attr_off, 1, putchar_x);
    break;
  case '':
    tputs(bold_mode_on, 1, putchar_x);
    tputs("_", 1, putchar_x);
    tputs(all_attr_off, 1, putchar_x);
    break;
  case '':
    tputs(bold_mode_on, 1, putchar_x);
    tputs("A", 1, putchar_x);
    tputs(all_attr_off, 1, putchar_x);
    break;
  default:
    putchar_x(ch);
  }
}

void print_input_buffer() {

  int i;

  if (buffer_length > cols - 10) {
    tputs(tgoto(cursor_motion, 0, status_line + 1), 1, putchar_x);
    if (input_cursor < cols - 10) {
      for (i=0; i<cols && i<buffer_length; i++) {
	t_putc(input_buffer[i]);
      }
      if (buffer_length < cols) {
	tputs(clear_to_eol, 1, putchar_x);
      }
      tputs(tgoto(cursor_motion, input_cursor, status_line + 1), 1, putchar_x);
    }
    else {
      for (i=input_cursor - (cols - 10); i- (input_cursor - (cols - 10)) < cols && i < buffer_length; i++) {
	t_putc(input_buffer[i]);
      }
      if (buffer_length - input_cursor < 10) {
	tputs(clear_to_eol, 1, putchar_x);
      }
      tputs(tgoto(cursor_motion, cols-10, status_line + 1), 1, putchar_x);
    }
  }
  else {
    tputs(tgoto(cursor_motion, 0, status_line + 1), 1, putchar_x);
    for (i=0; i<buffer_length; i++) {
      t_putc(input_buffer[i]);
    }
    tputs(clear_to_eol, 1, putchar_x);
    tputs(tgoto(cursor_motion, input_cursor, status_line + 1), 1, putchar_x);
  }
  fflush(stdout);
}


void restore_cursor() {

  if (buffer_length > cols - 10) {
    if (input_cursor < cols - 10) {
      tputs(tgoto(cursor_motion, input_cursor, status_line + 1), 1, putchar_x);
    }
    else {
      tputs(tgoto(cursor_motion, cols-10, status_line + 1), 1, putchar_x);
    }
  }
  else {
    tputs(tgoto(cursor_motion, input_cursor, status_line + 1), 1, putchar_x);
  }
  fflush(stdout);
}


void input_addchar(char ch) {

  char *attr_on, *attr_off;
  char outp_ch=' ';

  if (isprint(ch) || ch == '' || ch == '' || ch == '' || ch == '') {
    if (buffer_length > 1000) return;
    if (iscntrl(ch)) {
      attr_on  = bold_mode_on;
      attr_off = all_attr_off;
      switch (ch) {
      case '': outp_ch = 'B'; break;
      case '': outp_ch = 'V'; break;
      case '': outp_ch = '_'; break;
      case '': outp_ch = 'A'; break;
      }
    }
    else {
      attr_on  = empty_string;
      attr_off = empty_string;
      outp_ch = ch;
    }

    if (input_cursor == buffer_length) {
      input_buffer[input_cursor] = ch;
      input_buffer[input_cursor + 1] = '\0';
      input_cursor++;
      buffer_length++;
      if (buffer_length > cols - 10) {
	print_input_buffer();
      }
      else {
	tputs(attr_on, 1, putchar_x);
	putchar_x(outp_ch);
	tputs(attr_off, 1, putchar_x);
	fflush(stdout);
      }
    }
    else {
      memmove(&input_buffer[input_cursor + 1], &input_buffer[input_cursor], buffer_length - input_cursor + 1);
      input_buffer[input_cursor] = ch;
      input_cursor++;
      buffer_length++;
      print_input_buffer();
    }
  }
  else {
    switch (ch) {
    case '\r': case '\n':
      add_history(input_buffer);
      if (!on_console(input_buffer)) {
	switch (console_mode()) {
	case CONSOLE_IRC_MODE:
	  parse_irc_console(input_buffer);
	  break;
	case CONSOLE_PARTY_MODE:
	  parse_party_console(input_buffer);
	  break;
	case CONSOLE_COMMAND_MODE:
	  parse_cmd_console(input_buffer);
	  break;
	}
      }
      strcpy(input_buffer, "");
      buffer_length = 0;
      input_cursor = 0;
      tputs(tgoto(cursor_motion, 0, status_line + 1), 1, putchar_x);
      tputs(clear_to_eol, 1, putchar_x);
      fflush(stdout);
      break;
    case '\t':
      if (console_mode() == CONSOLE_IRC_MODE) {
	add_history(input_buffer);
	sprintf(input_buffer, "/msg %s ", msg_history_prev());
	buffer_length = input_cursor = strlen(input_buffer);
	print_input_buffer();
      }
      break;
    case '\b':
      if (input_cursor > 0) {
	memmove(&input_buffer[input_cursor - 1], &input_buffer[input_cursor], buffer_length - input_cursor + 1);
	input_cursor--;
	buffer_length--;
	print_input_buffer();
      }
      break;
    case '':
      term_print_mode_line();
      term_print_status_line();
      term_flush_window(TERM_SNOOP_WINDOW);
      term_flush_window(TERM_IRC_WINDOW);
      print_input_buffer();
      break;
    case '':
      switch (console_mode()) {
      case CONSOLE_IRC_MODE:
	console_setmode(CONSOLE_COMMAND_MODE);
	break;
      case CONSOLE_COMMAND_MODE:
	console_setmode(CONSOLE_PARTY_MODE);
	break;
      case CONSOLE_PARTY_MODE:
	console_setmode(CONSOLE_IRC_MODE);
	break;
      }
      break;
    }
  }
}


void parse_key_left() {

  if (input_cursor == 0) return;

  input_cursor--;

  if (buffer_length < cols - 10)
    restore_cursor();
  else
    print_input_buffer();
}


void parse_key_right() {

  if (input_cursor == buffer_length) return;

  input_cursor++;

  if (buffer_length < cols - 10)
    restore_cursor();
  else
    print_input_buffer();
}


void parse_key_up() {

  strcpy(input_buffer, history_prev());
  buffer_length = input_cursor = strlen(input_buffer);
  print_input_buffer();
}

void parse_key_down() {
  
  strcpy(input_buffer, history_next());
  buffer_length = input_cursor = strlen(input_buffer);
  print_input_buffer();
}


void read_and_parse_terminal() {

  static char buffer[10];
  static int  b_ptr=0;
  char input_ch;

  read(fileno(stdin), &input_ch, 1);

  if (b_ptr == 0) {
    if (input_ch == key_up_arrow[0]) {
      buffer[0] = input_ch;
      b_ptr = 1;
    }
    else {
      input_addchar(input_ch);
    }
  }
  else {
    buffer[b_ptr]   = input_ch;
    buffer[b_ptr+1] = '\0';
    b_ptr++;
    if (b_ptr > strlen(key_up_arrow)) {
      b_ptr = 0;
      input_addchar(input_ch);
    }
    else {
      if (b_ptr == strlen(key_up_arrow)) {
	if (key_left_arrow[b_ptr - 1] == buffer[b_ptr - 1]) {
	  parse_key_left();
	}
	else if (key_right_arrow[b_ptr - 1] == buffer[b_ptr - 1]) {
	  parse_key_right();
	}
	if (key_up_arrow[b_ptr - 1] == buffer[b_ptr - 1]) {
	  parse_key_up();
	}
	else if (key_down_arrow[b_ptr - 1] == buffer[b_ptr - 1]) {
	  parse_key_down();
	}
	b_ptr = 0;
      }
    }
  }
}
